package com.schbrain.canal.client.core;

import com.schbrain.canal.client.event.ResolverCanalEvent;
import com.schbrain.canal.client.exception.CanalClientException;
import com.schbrain.canal.client.exception.ReflectionException;
import com.schbrain.canal.client.reflector.DefaultReflectorFactory;
import com.schbrain.canal.client.reflector.Reflector;
import com.schbrain.canal.client.reflector.ReflectorFactory;
import com.schbrain.canal.client.reflector.DefaultObjectFactory;
import com.schbrain.canal.client.transfer.ObjectFactory;
import com.schbrain.canal.client.type.TypeHandler;
import com.schbrain.canal.client.type.TypeHandlerRegister;
import com.schbrain.canal.client.utils.Dml;
import com.schbrain.canal.client.utils.MapUnderscoreToCamelCase;
import lombok.extern.slf4j.Slf4j;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.text.ParseException;
import java.util.*;

/**
 * @author zhuyf
 * @date 2022/6/21
 */
@Slf4j
public class MethodArgumentResolver {

    public static Map<ResolverCanalEvent, Class> classNameMappingClass = new HashMap<>();

    private ObjectFactory objectFactory = new DefaultObjectFactory();

    private ReflectorFactory reflectorFactory = new DefaultReflectorFactory();

    /**
     * 获取参数类
     * @param event
     * @return
     */
    public Class getArgumentClass(ResolverCanalEvent<?> event) {
        if (!classNameMappingClass.containsKey(event)) {
            //接口泛型缓存
            boolean cache = interfaceGeneric(event);
            if(!cache){
                //继承类泛型缓存
                cache  = extendGeneric(event);
            }
            if(!cache){
                CanalClientException e = new CanalClientException(event.getClass() + "generic get exception");
                log.error("generic get exception",e);
                throw e;
            }
        }
        return classNameMappingClass.get(event);
    }


    /**
     * 接口泛型缓存
     * @param event
     * @return
     */
    private boolean interfaceGeneric(ResolverCanalEvent<?> event){
        Type[] types = event.getClass().getGenericInterfaces();
        if(types == null || types.length == 0){
            return false;
        }
        for (Type type : types) {
            if(!(type instanceof ParameterizedType)){
                continue;
            }
            ParameterizedType parameterized  = (ParameterizedType)type;
            Type rawType = parameterized.getRawType();
            if(rawType.equals(ResolverCanalEvent.class)){
                Class clazz = (Class)parameterized.getActualTypeArguments()[0];
                classNameMappingClass.put(event, clazz);
                return true;
            }
        }
        return false;
    }

    private boolean extendGeneric(ResolverCanalEvent<?> event){
        Class c = event.getClass();
        while(true){
            Type type  = c.getGenericSuperclass();
            if(type instanceof ParameterizedType){
                ParameterizedType parameterizedType = (ParameterizedType)type;
                Type[] types = parameterizedType.getActualTypeArguments();
                if(types!= null && types.length>0){
                    classNameMappingClass.put(event, (Class) types[0]);
                    return true;
                }
            }
            c = c.getSuperclass();
            if(type.equals(Object.class)){
               break;
            }
        }
        return false;
    }

    public List<EditMetaInfo> resolver(ResolverCanalEvent event, Dml dml) throws InvocationTargetException, IllegalAccessException, ParseException,ReflectionException {
        Class c  = getArgumentClass(event);
        List<EditMetaInfo> editMetaInfos=new ArrayList<>();
        List<Dml.Row> datas = dml.getData();
        for (Dml.Row row : datas) {
            EditMetaInfo metaInfo = new EditMetaInfo();
            Object after = columnsConvertObject(c,row.getData());
            metaInfo.setAfter(after);
            Object before = columnsConvertObject(c,row.getOld());
            metaInfo.setBefore(before);
            editMetaInfos.add(metaInfo);
        }
        return editMetaInfos;
    }


    public Object columnsConvertObject(Class c,Map<String,Object> columns) throws ReflectionException, InvocationTargetException, IllegalAccessException, ParseException {
        if(columns==null||columns.isEmpty()){
            return null;
        }
        Reflector classesReflector= reflectorFactory.findForClass(c);
        Object o = objectFactory.create(c);
        for (String columnName: columns.keySet()) {
            String filedName= MapUnderscoreToCamelCase.convertByCache(columnName);
            Object value=columns.get(columnName);
            if(value==null){
                continue;
            }
            if (!classesReflector.hasGetter(filedName)) {
                continue;
            }
            Type setterType= classesReflector.getSetterType(filedName);
            TypeHandler typeHandler= TypeHandlerRegister.getTypeHandler(setterType);
            if(typeHandler==null){
                log.error("未适配到typeHandle{},name:{},value:{},",setterType,filedName,value);
            }
            classesReflector.getSetInvoker(filedName).invoke(o,new Object[]{typeHandler.convert(value)});
        }
        return o;
    }

}
